#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include "charclass.h"
#include "classmap.h"



static void lSetCharMap( Mel *map, CharRange range[] )
{
	unsigned	i;

	
	for (i=0; range[i].start != 0; ++i)
	{
 		unsigned j;

 		j=range[i].start;
 		do
 			SETBIT( map, j );
    	while ( j++ < range[i].end);
	}
}


#define ONES	1
#define ZEROS	0 

static void lInitCompressedMap( Mel cmap[], int *max )
{
	memset(&cmap[0], 0, BYTES_PER_PAGE);
	memset(&cmap[MELS_PER_PAGE], 0xff, BYTES_PER_PAGE);
	*max = 2;
}

static void lMakeCompressedMap( PageMap page[], Mel cmap[], int *max, Mel map[] )
{
	int	i;

	
	for (i=0; i < PAGE_MAX; ++i)
	{
		Mel	first=map[MELS_PER_PAGE*i];
		int	j=0;

		if ( first == 0 || first == (Mel)~0 )
		{
			for (; j < MELS_PER_PAGE; ++j)
			{
				if ( map[MELS_PER_PAGE*i+j] != first )
					break;
			}
		}

		if ( j == MELS_PER_PAGE )	
		{
			if ( first == 0 )
				page[i] = ZEROS;
			else if ( first == (Mel)~0 )
				page[i] = ONES;
		}
		else
		{
			memcpy( &cmap[*max * MELS_PER_PAGE], &map[i*MELS_PER_PAGE], BYTES_PER_PAGE);
			page[i] = *max;
			++*max;
		}

	}
}


static void lGeneratePageArray( PageMap page[PAGE_MAX], char const *aPageName )
{
	int i;

	printf("const PageMap %s[%d] = {\n", aPageName, PAGE_MAX );

	for (i=0; i < PAGE_MAX; )
	{
		int	j=0;

		printf("\t");
		for (j=0; i < PAGE_MAX && j < 80 / 4 - 8; ++i, ++j )
			printf("%3d, ", page[i]);
		printf("\n");
	}
	printf("};\n");
}

static void lGenerateMapArray( Mel map[], const char *aMapName, int aPageMax )
{
	int i;

	printf("Mel const %s[%d] = {\n", aMapName, aPageMax * MELS_PER_PAGE);

	for (i=0; i < aPageMax * MELS_PER_PAGE; )
	{
		int	j=0;

		printf("\t");
		for (j=0; i < aPageMax * MELS_PER_PAGE && j < (80-8) / (5 + BITS_PER_MEL / 4 ); ++i, ++j )
			printf("0x%0*x, ", BITS_PER_MEL / 4, map[i]);
		printf("\n");
	}
	printf("};\n");
}

int main(void)
{
	Mel		map[CHARMAP_MAX];
	PageMap	page[PAGE_MAX];
	Mel		cmap[CHARMAP_MAX];
	int		page_max;

	/* Initialise the global cmap */
	lInitCompressedMap(cmap, &page_max);

	/* Letter ::= BaseChar ,  Ideographic */

	printf("#include \"classmap.h\"\n\n");

	/* BaseChar */
	memset( map, 0, CHARPMAP_SIZE );
	lSetCharMap ( map, BaseChar );
	lMakeCompressedMap( page, cmap, &page_max, map );
	lGeneratePageArray( page, "BaseChar_Page" );

	/* Ideographic */
	memset( map, 0, CHARPMAP_SIZE );
	lSetCharMap ( map, Ideographic );
	lMakeCompressedMap( page, cmap, &page_max, map );
	lGeneratePageArray( page, "Ideographic_Page" );

	/* CombiningChar */
	memset( map, 0, CHARPMAP_SIZE );
	lSetCharMap ( map, CombiningChar );
	lMakeCompressedMap( page, cmap, &page_max, map );
	lGeneratePageArray( page, "CombiningChar_Page" );

	/* Digit */
	memset( map, 0, CHARPMAP_SIZE );
	lSetCharMap ( map, Digit );
	lMakeCompressedMap( page, cmap, &page_max, map );
	lGeneratePageArray( page, "Digit_Page" );

	/* Extender */
	memset( map, 0, CHARPMAP_SIZE );
	lSetCharMap ( map, Extender );
	lMakeCompressedMap( page, cmap, &page_max, map );
	lGeneratePageArray( page, "Extender_Page" );

	/* Output the cmap */
	lGenerateMapArray( cmap, "Shared_Map", page_max );

	return EXIT_SUCCESS;
}
