#include <stdlib.h>
#include <stddef.h>
#include <string.h>


#include "DStr.h"



/* 
** A Dymanic string module, or 'DStr'
** Common idiom, using struct member chumminess to allow for flat contigous structure, and the imitation of
** traditional zero terminated strings. Valid in ANSI C99, non-conforming to C89, but all compilers appear to accept it.
*/


static const int GrowFactor = 2;

#define	DSTR_INITIAL_SIZE	10

int DStr_Create( DStr** ds )
{
	*ds = malloc( sizeof (DStr) + DSTR_INITIAL_SIZE * sizeof (XMLChar) );
	if ( *ds )
	{
		DStr *d = *ds;

		d->size = DSTR_INITIAL_SIZE;
		d->count = 0;
		d->s[0]='\0';
	}
	return ds == NULL;
}

int DStr_EnsureCapacity( DStr **ds, size_t capacity )
{
	int		ret = 0;
	DStr	*d = *ds;
	
	if ( d->size < capacity )
	{
		DStr*	new_ds;
		capacity *= GrowFactor;	/* Put in the grow factor */
		new_ds = realloc( d, sizeof *d + capacity * sizeof (XMLChar) );
		if ( new_ds )
		{
			new_ds->size=capacity;
			*ds = new_ds;
		}
		else
			ret = 1;
	}
	return ret;
}

int DStr_EnsureRemaining( DStr **ds, size_t remaining )
{
	DStr	*d = *ds;
	size_t	capacity = remaining + d->count;

	return DStr_EnsureCapacity( ds, capacity );
}


int	DStr_AppendChar( DStr** ds, XMLChar c )
{
	int		ret = 0;

	ret = DStr_EnsureRemaining( ds, 2);
	
	if ( ! ret )
	{
		DStr *d = *ds;
		d->s[ d->count++ ] = c;
		d->s[ d->count ] = '\0';
	}
	return ret;
}

int	DStr_AppendCStr( DStr** ds, XMLChar* s )
{
	int		ret = 0;
	size_t	slen = XML_STRLEN(s);

	ret = DStr_EnsureRemaining( ds, slen + 1 );
	
	if ( ! ret )
	{
		DStr *d = *ds;
		XML_STRCPY(&d->s[ d->count ], s);
		d->count += slen;
		d->s[ d->count ] = '\0';
	}
	return ret;
}



size_t DStr_Length( DStr* ds )
{
	return ds->count;
}

XMLChar	*DStr_CStr( DStr *ds )
{
	return ds->s;
}

int DStr_Destroy( DStr** ds )
{
	int	ret = 0;
	if ( *ds )
	{
		free( *ds ); *ds = NULL; 
	}
	else
	{
		ret = 1;
	}
	return ret;
}