#include <stdlib.h>
#include <memory.h>
#include <assert.h>
#include "LexBuffer.h"
#include "DStr.h"

LexBuffer *LexBuffer_Create(void)
{
	LexBuffer*	s = malloc( sizeof (LexBuffer) );
	if ( s )
	{
		LexBlock*	sb;
		memset( s, 0, sizeof (LexBuffer) );
		s->startBlock = malloc( sizeof (LexBlock));
		if ( s->startBlock )
		{
			sb = s->startBlock;
			sb->n = 0;
			sb->next = NULL;
			sb->prev = NULL;
			s->forwardBlock = s->startBlock;
			s->endBlock = s->startBlock;
			s->gap = 0;
			s->lineNo=1;
			s->colNo=1;
			s->tabSize=LEX_TAB_WIDTH;
		}
		else
		{
			free( s ); s = NULL;
		}

	}
	return s;
}

int	LexBuffer_Destroy(LexBuffer* source)
{
	LexBlock* block;

	block = source->startBlock;
	while ( block )
	{
		LexBlock*	next = block->next;
		free( block );
		block = next;
	}
	free( source );
	return 0;
}

int	LexBuffer_SetTabSize(LexBuffer* l, size_t tab_size)
{
	l->tabSize = tab_size;
	return 0;
}

int	LexBuffer_AddData(LexBuffer* source, XMLChar* data, size_t n)
{
	LexBlock* sb;
	size_t	i = 0;

	sb = source->endBlock;

	while ( i < n ) 
	{
		if ( sb->n == LEX_BLOCK_SIZE )
		{
			LexBlock	*sbNew = malloc( sizeof (LexBlock) );
			if ( !sbNew )
				return 1;
			sb->next = sbNew;

			
			sbNew->n = 0;
			sbNew->next = NULL;
			sbNew->prev = sb;
			source->endBlock = sbNew;
			sb = sbNew;
		}
		sb->d[ sb->n++ ] = data[i];
		++i;
	}	
	return 0;
}


tInt LexBuffer_NextChar(LexBuffer* s)
{
	LexBlock* sb;


	sb = s->forwardBlock;
	if ( s->forward == sb->n )
	{
		
		if ( s->forwardBlock->next )
		{
			s->forwardBlock = s->forwardBlock->next;
			s->forward = 0;
			sb = s->forwardBlock;

		}
		else
			return -1;
	}

	s->gap++;
	return sb->d[ s->forward++ ];
}

tInt LexBuffer_PrevChar(LexBuffer* s)
{
	LexBlock* sb;

	sb = s->forwardBlock;
	if ( s->forward == 0 )
	{
		
		if ( s->forwardBlock->prev )
		{
			s->forwardBlock = s->forwardBlock->prev;
			s->forward = s->forwardBlock->n;
			sb = s->forwardBlock;

		}
		else
			return -1;
	}

	s->gap--;
	return sb->d[ --s->forward ];
}

int LexBuffer_Move(LexBuffer *lex, int delta)
{
	int	ret=0, sign;
	
	sign=delta < 0 ? -1 : delta > 0 ? 1 : 0;

	if ( sign > 0 )
	{
		while ( ret != delta && LexBuffer_NextChar(lex) != -1  )
		{
			ret++;
		}
	}
	else if ( sign < 0 )
	{
		while ( ret != delta && LexBuffer_PrevChar(lex) != -1  )
		{
			ret--;
		}
	}
	return ret;
}

static void sCountLines(LexBuffer* s)
{
	LexBlock *sb = s->startBlock;
	int i;
	int j = s->start;

	for (i = 0; i < s->gap; ++i)
	{
		if ( j == sb->n )
		{
			sb = sb->next;
			j = 0;
		}
		if ( sb->d[j] == '\t')
			s->colNo+=(unsigned long)s->tabSize;
		else
			s->colNo++;
		if ( sb->d[ j++ ] == '\n')
		{
			s->lineNo++;
			s->colNo=1;
		}
	}
}

int	LexBuffer_Accept(LexBuffer* s)
{
	LexBlock *sb, *next;

	sCountLines(s);

	sb = s->startBlock;
	while ( sb != s->forwardBlock )
	{
		next = sb->next;
		free(sb);
		sb = next;
		sb->prev = NULL;
	}

	s->start = s->forward;
	s->startBlock = s->forwardBlock;
	s->gap = 0;
	return 0;
}


int	LexBuffer_Rewind(LexBuffer* s)
{
	s->forward = s->start;
	s->forwardBlock = s->startBlock;
	s->gap = 0;
	return 0;
}

int LexBuffer_Copy( LexBuffer *l, DStr **dest )
{
	DStr_Create( dest );
	return LexBuffer_Append(l, dest );
}

int LexBuffer_Append(LexBuffer *s, DStr **dest)
{
	int			i;
	LexBlock	*sb;
	int			j;
	int			ret = 0;

	if ( s->gap > 0 )
	{
		if ( *dest )
		{
			XMLChar	*dp;

			DStr_EnsureRemaining( dest, s->gap + 1 );	/* Prepare the target area */
			dp = (*dest)->s+(*dest)->count;				/* Get the target area */
			(*dest)->count+= s->gap;					/* Set count now */	
		
			sb = s->startBlock;
			j = s->start;
			for (i = 0; i < s->gap; ++i)
			{
				if ( j == sb->n )
				{
					sb = sb->next;
					j = 0;
				}
				*dp++ = sb->d[ j++ ];
			}
			*dp = '\0';		/* Null terminate now */
		}
		else
		{
			ret = 1;
		}
	}
	else
	{
		ret = 1;
	}
	return ret;
	
}

int LexBuffer_IsEmpty(LexBuffer *s)
{
	return s->forwardBlock == s->endBlock && s->forward == s->endBlock->n;
}

int	LexBuffer_Gap(LexBuffer *p)
{
	return p->gap;
}