#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>


#include "xml.h"
#include "DStr.h"
#include "classmap.h"

static XMLError	TextInit( XML *x, StateRecord *sr );
static XMLError TextRun( XML *, StateRecord * );
static XMLError AmpersandRun( XML *, StateRecord * );
static XMLError AmpersandDone( XML *, StateRecord * );
static XMLError TagRun( XML *, StateRecord * );
static XMLError StartTagInit( XML *, StateRecord * );
static XMLError StartTagRun( XML *, StateRecord * );
static XMLError StartTagDone( XML *, StateRecord * );
static XMLError EndTagRun( XML * x, StateRecord * );
static XMLError EndTagDone( XML * x, StateRecord * );
static XMLError IdentifierRun( XML *, StateRecord * );
static XMLError AttributeInit( XML *, StateRecord * );
static XMLError AttributeRun( XML *, StateRecord * );
static XMLError AttributeDone( XML *, StateRecord * );
static XMLError StringRun( XML *, StateRecord * );
static XMLError CharacterReferenceRun( XML *, StateRecord *);
static XMLError WhiteRun( XML *, StateRecord * );
static XMLError	CommentRun( XML *, StateRecord * );
static XMLError	DoctypeRun( XML *, StateRecord * );
static XMLError	NameAndAttributeRun(XML *, StateRecord *);
static XMLError PITagInit( XML *, StateRecord *);
static XMLError	PITagRun( XML *, StateRecord * );
static XMLError PITagDone( XML *, StateRecord * );
static XMLError	CDATARun( XML *, StateRecord * );

ScannerDescriptor	TextScan =					{ TextInit,			TextRun,					NULL };
ScannerDescriptor	AmpersandScan =				{ NULL,				AmpersandRun,				AmpersandDone };
ScannerDescriptor	TagScan =					{ NULL,				TagRun,						NULL };
ScannerDescriptor	StartTagScan =				{ StartTagInit,		StartTagRun,				StartTagDone };
ScannerDescriptor	EndTagScan =				{ NULL,				EndTagRun,					EndTagDone };
ScannerDescriptor	IdentifierScan =			{ NULL,				IdentifierRun,				NULL };
ScannerDescriptor	AttributeScan =				{ AttributeInit,	AttributeRun,				AttributeDone };
ScannerDescriptor	StringScan =				{ NULL,				StringRun,					NULL };
ScannerDescriptor	CharacterReferenceScan =	{ NULL,				CharacterReferenceRun,		NULL };
ScannerDescriptor	WhiteScan =					{ NULL,				WhiteRun,					NULL };
ScannerDescriptor	CommentScan =				{ NULL,				CommentRun,					NULL } ;
ScannerDescriptor	DoctypeScan =				{ NULL,				DoctypeRun,					NULL };
ScannerDescriptor	PITagScan =					{ StartTagInit,		PITagRun,					StartTagDone };
ScannerDescriptor	NameAndAttributeScan =		{ NULL,				NameAndAttributeRun,		NULL };
ScannerDescriptor	CDATAScan =					{ NULL,				CDATARun,					NULL };

/* (void)0 is to prevent the macro from being avaluated in an expression */
#define NOTIFY_TEXT( x, text ) \
	( (x->TextNotify) ? (x)->TextNotify(x->UserData, text ) : (void)0 )
#define NOTIFY_START_TAG(x, name, attribs) \
	( (x->StartTagNotify) ? (x)->StartTagNotify(x->UserData, name, attribs) : (void)0 )
#define	NOTIFY_END_TAG(x, name)	\
	( (x->EndTagNotify) ? (x)->EndTagNotify(x->UserData, name) : (void)0 )
#define NOTIFY_COMMENT(x, comment) \
	( (x->CommentNotify) ? (x)->CommentNotify(x->UserData, comment) : (void)0 )
#define NOTIFY_PI_TAG(x, name, attribs) \
	( (x->PITagNotify) ? (x)->PITagNotify(x->UserData, name, attribs) : (void)0 )
#define NOTIFY_PARSE_ERROR(x, parse_err)	\
	( (x->ParseErrorNotify) ? \
	(x)->ParseErrorNotify(x->UserData, x->LexBuffer->lineNo, x->LexBuffer->colNo, parse_err) : (void)0 )

/* Text scan
*****************************************************************************************************/
#define TEXTCOUNT	(sr->local1.t_usigned_long)
#define ENTITY_REP	(sr->local2.t_pDStr)

static XMLError	TextInit( XML *x, StateRecord *sr )
{
	TEXTCOUNT=0;
	return eXML_OK;
}

static XMLError TextRun( XML *x, StateRecord *sr )
{

	DStr			*ds;
	XMLError		result = eXML_OK;
	int				c;
	enum			{TEXT, ENTITY_REF};

	switch (sr->state)
	{
	case TEXT:	/* Text parsing */
        c = LexBuffer_NextChar(x->LexBuffer);
		while ( c != '<' && c!='&' && c != -1 && TEXTCOUNT < x->options.MaxTextChunk-1)
		{
			TEXTCOUNT++;
			c = LexBuffer_NextChar(x->LexBuffer);
		}
		
		if ( c!=-1 )
		{
			if ( c=='&' )	/* Character reference, or entity reference */
			{
				Parameter p1; p1.t_ppDStr = &ENTITY_REP;
				LexBuffer_PrevChar(x->LexBuffer);
				DStr_Create( &ENTITY_REP );
				SM_Call1(x, &AmpersandScan, &p1);
				sr->state = ENTITY_REF;
			}
			else if ( c=='<' ) /* A tag */
			{
				LexBuffer_PrevChar(x->LexBuffer);
				SM_Call( x, &TagScan ); /* Go parse the tag now */
			}

			/* Collect text up until now and report it */
			LexBuffer_Copy(x->LexBuffer, &ds);
			LexBuffer_Accept(x->LexBuffer);
			if ( ds )
			{
				NOTIFY_TEXT( x, ds );
				DStr_Destroy(&ds);
			}
			TEXTCOUNT=0;
		}
		break;

	case ENTITY_REF:
		if ( ENTITY_REP )
		{
			NOTIFY_TEXT( x, ENTITY_REP );
			DStr_Destroy(&ENTITY_REP);
		}
		sr->state = TEXT;
	}
	return result;

}
#undef TEXTCOUNT
#undef ENTITY_REP
/* Entity scan 
**************************************************************************************************/

const struct AmpRec
{
	const XMLChar * const	name;
	XMLChar					val;
} AmpTable[] = { 
	XML_TEXT("amp"), '&',
	XML_TEXT("gt"), '>',
	XML_TEXT("lt"), '<',
	XML_TEXT("apos"), '\'',
	XML_TEXT("quot"), '"',
	NULL, 0
};

static XMLError AmpersandRun( XML *x, StateRecord *sr )
{
#define CHAR_REP	(sr->param1.t_ppDStr)
	XMLError	result = eXML_OK;
	int	c;
	enum { AMP, DETERMINE_TYPE, NAMED, VALUED, MAKE_VALUED };

	switch (sr->state)
	{
	case AMP:
		c = LexBuffer_NextChar(x->LexBuffer);
		if (c!=-1)
		{
			assert( c=='&' );
			LexBuffer_Accept(x->LexBuffer);
			sr->state=DETERMINE_TYPE;
		}
		break;

	case DETERMINE_TYPE:
		c = LexBuffer_NextChar(x->LexBuffer);
		if (c!=-1)
		{
			if ( c=='#' )
			{
				sr->state = VALUED;
				LexBuffer_Accept(x->LexBuffer);
			}
			else /* Assume named */
			{
				LexBuffer_PrevChar(x->LexBuffer);
				sr->state = NAMED;
			}
		}
		break;

	case NAMED:
		c = LexBuffer_NextChar(x->LexBuffer);
		if (c==-1)
			return result;
		if ( c==';' )
		{
			DStr	*dest;
			struct AmpRec const *a = AmpTable;
			LexBuffer_PrevChar(x->LexBuffer);
			LexBuffer_Copy(x->LexBuffer, &dest);
			if ( dest )
			{
				while ( a->name )
				{
					if ( XML_STRCMP( a->name, DStr_CStr(dest) ) == 0 )
						break;
					a++;
				}
				if ( a->name ) /* Found it ? */
				{
					if ( CHAR_REP )
					{
						DStr_AppendChar( CHAR_REP, a->val );
					}
					else
					{
						result = eXML_OutOfMemory;
					}
				}
				else
				{
					NOTIFY_PARSE_ERROR(x, eXMLParseErr_UndefinedEntityID);
				}
				DStr_Destroy(&dest);
			}
			else
			{
				NOTIFY_PARSE_ERROR(x, eXMLParseErr_MalformedEntity);
			}
			LexBuffer_NextChar(x->LexBuffer);
			LexBuffer_Accept(x->LexBuffer);
			SM_Return(x);
		}
		break;

	case VALUED:
		c = LexBuffer_NextChar(x->LexBuffer);
		if (c!=-1)
		{
			if ( c=='x' ) /* hex */
			{
				Parameter	p1, p2;
				p1.t_ppDStr = CHAR_REP;
				p2.t_int=16;
				SM_Call2(x, &CharacterReferenceScan, &p1, &p2); /* Include base */
				LexBuffer_Accept(x->LexBuffer);
			}
			else /* Decimal */
			{
				Parameter	p1, p2;
				p1.t_ppDStr = CHAR_REP;
				p2.t_int=10;
				SM_Call2(x, &CharacterReferenceScan, &p1, &p2); /* Include base */
				LexBuffer_PrevChar(x->LexBuffer);
			}
		}
		sr->state = MAKE_VALUED;
		break;

	case MAKE_VALUED:
		c = LexBuffer_NextChar(x->LexBuffer);
		if ( c!=-1 )
		{
			if ( c==';' )
			{
				LexBuffer_Accept(x->LexBuffer);
			}
			SM_Return(x);
		}
		break;

	}
	return result;
#undef CHAR_REP
}

static XMLError AmpersandDone( XML * x, StateRecord *sr )
{
	if ( sr->work )
		free(sr->work);
	return eXML_OK;
}

/* Tag (general) scan 
**************************************************************************************************/
static XMLError TagRun( XML * x, StateRecord *sr )
{
	XMLError	result = eXML_OK;
	enum TagStates { TAG_OPEN, TAG_TYPE, BANG_TYPE, END_WHITE, TAG_CLOSE, SKIP_TO_CLOSE };
	int	c;

	switch ( sr->state )
	{
	case TAG_OPEN:
		c = LexBuffer_NextChar(x->LexBuffer);
		if (c==-1)
			return result;
		assert( c == '<' ); /* Must be a start tag to get here */
		LexBuffer_Accept(x->LexBuffer);
		sr->state = TAG_TYPE;
		break;

	case TAG_TYPE:
		c = LexBuffer_NextChar(x->LexBuffer);
		if (c==-1)
			return result;
		sr->state = END_WHITE;
		if ( c == '/' ) /* An end tag */
		{
			LexBuffer_Accept(x->LexBuffer);
			SM_Call(x, &EndTagScan );
		}
		else if ( c == '?' ) /* A PI tag */
		{
			LexBuffer_Accept(x->LexBuffer);
			SM_Call(x, &PITagScan );	/* Handle like normal tag for now */
		}
		else if ( c == '!' ) /* A bang, or exclamation mark - find out what sort of bang it is */
		{
			sr->state=BANG_TYPE;
		}
		else /* Then it should be a valid start tag */
		{
			LexBuffer_PrevChar(x->LexBuffer); /* Put back the first character of the name */
			SM_Call(x, &StartTagScan );			
		}
		break;

	case BANG_TYPE:
		c = LexBuffer_NextChar(x->LexBuffer);
		if (c==-1)
			return result;
		if ( c=='-' )	/* A comment ? */
		{
			SM_Call(x, &CommentScan );
		}
		else if ( c == '[' ) /* A CDATA section ? */
		{
			SM_Call(x, &CDATAScan);
		}
		else if ( c == 'D' ) /* Then it's most probably a prolog like DOCTYPE */
		{
			LexBuffer_PrevChar(x->LexBuffer);
			SM_Call(x, &DoctypeScan);
		}
		sr->state=END_WHITE;
		break;

	case END_WHITE:
		SM_Call(x,&WhiteScan);
		sr->state=TAG_CLOSE;
		break;

	case TAG_CLOSE:
		c = LexBuffer_NextChar(x->LexBuffer);
		if (c!=-1)
		{
			if ( c=='>' )
			{
				LexBuffer_Accept(x->LexBuffer);
				SM_Return(x);
			}
			else
			{
				NOTIFY_PARSE_ERROR(x, eXMLParseErr_SpuriousCharacters );
				sr->state=SKIP_TO_CLOSE;
			}
		}
		break;

	case SKIP_TO_CLOSE: /* Recover, searchinf for > */
		do
		c = LexBuffer_NextChar(x->LexBuffer);
		while (c!=-1 && c!='>');
		if ( c!= -1 )
		{
			LexBuffer_Accept(x->LexBuffer);
			SM_Return(x);
		}
		break;
	}
	return result;
}


/* Name and attribute scan (p1=&name, p2=&attribs);
**************************************************************************************************/
static XMLError NameAndAttributeRun(XML *x, StateRecord *sr)
{
	enum { TAG_NAME, CHECK_NAME, ATTRIBUTE, ACCEPT }; 

	switch (sr->state)
	{
	case TAG_NAME:
		{
			sr->state = CHECK_NAME;
			SM_Call1( x, &IdentifierScan, &sr->param1 );
		}
		break;

	case CHECK_NAME:
		if ( *sr->param1.t_ppDStr )
			sr->state = ATTRIBUTE;
		else /* Parse error */
		{
			NOTIFY_PARSE_ERROR(x, eXMLParseErr_IdentifierMalformed);
			SM_Return(x);
		}
		break;

	case ATTRIBUTE:
		{
			sr->state = ACCEPT;
			SM_Call1(x, &AttributeScan, &sr->param2);
		}
		break;

	case ACCEPT:
		SM_Return(x);
		break;
	}
	return	eXML_OK;
}

/* StartTag scan
**************************************************************************************************/

typedef struct
{
	DStr		*name;
	AttribList	*attribs;
} TagWork;

static XMLError StartTagInit(XML * x, StateRecord *sr )
{
	sr->work = malloc( sizeof (TagWork) );
	if ( sr->work )
	{
		TagWork	*w = sr->work;
		w->name=NULL;
		w->attribs=NULL;
	}
	else
		return eXML_OutOfMemory; /* memory */
	return eXML_OK;
}


static XMLError StartTagRun( XML * x, StateRecord *sr )
{
	XMLError	result = eXML_OK;
	enum TagStates { NAME_AND_ATTRIBUTE, NOTIFY, WHITE, EMPTY_TAG };
	int	c;
	TagWork	*w = sr->work;

	switch ( sr->state )
	{
	case NAME_AND_ATTRIBUTE:
		{
			Parameter p1, p2;
			p1.t_ppDStr=&w->name; p2.t_ppvoid=&w->attribs;
			SM_Call2(x, &NameAndAttributeScan, &p1, &p2 );
			sr->state=NOTIFY;
		}
		break;

	case NOTIFY:	/* Send a token out */
		if ( w->name )
			NOTIFY_START_TAG( x, w->name, w->attribs );	
		sr->state=WHITE;
		break;

	case WHITE:
		SM_Call(x, &WhiteScan );
		sr->state=EMPTY_TAG;
		break;

	case EMPTY_TAG: /* Scan for '/' at end of tag */
		c = LexBuffer_NextChar(x->LexBuffer);
		if ( c!=-1 )
		{
			if ( c == '/' )
			{
				NOTIFY_END_TAG(x, w->name );
				LexBuffer_Accept(x->LexBuffer);
			}
			else
			{
				LexBuffer_PrevChar(x->LexBuffer);
			}
			SM_Return(x);
		}
		break;

	}
	return result;
}

static XMLError StartTagDone( XML *x, StateRecord *sr )
{
	if ( sr->work )
	{
		TagWork	*w = sr->work;

		if ( w->name )
			free(w->name);
		if ( w->attribs )
		{
			AttribList	*al = w->attribs;
			do
			{
				AttribList	*temp = al->next;
				free(al->name);
				free(al->value);
				free(al);
				al=temp;
			} while (al);
		}
		free( sr->work );
	}
	return eXML_OK;
}

/* End tag scan
*******************************************************************************/

static XMLError EndTagRun( XML * x, StateRecord *sr )
{
	enum { GET_ID, NOTIFY };

	switch (sr->state)
	{
	case GET_ID:
		{
			Parameter	p; p.t_ppvoid = &sr->work;
			SM_Call1(x, &IdentifierScan, &p );
			sr->state = NOTIFY;
		}
		break;

	case NOTIFY:
		if ( sr->work ) /* was an identifier found ? */
		{
			NOTIFY_END_TAG(x, sr->work );
		}
		else
		{
			NOTIFY_PARSE_ERROR(x, eXMLParseErr_IdentifierMalformed);
		}
		SM_Return(x);
		break;
	}
	return eXML_OK;
}

static XMLError EndTagDone( XML * x, StateRecord *sr )
{
	if ( sr->work )
		free(sr->work);
	return eXML_OK;
}

/* Identifier parse
*********************************************************************************/
/*
**
** [4] NameChar ::= Letter | Digit | '.' | '-' | '_' | ':' | CombiningChar | Extender
** [5] Name ::= (Letter | '_' | ':') (NameChar)*
** [84] Letter ::= BaseChar | Ideographic
*/
#define	Letter(c)		( XML_ISBaseChar(c) || XML_ISIdeogaphic(c) )
#define NameChar(c)		( Letter(c) || XML_ISDigit(c) || c=='.' || c=='-' ||c=='_' || c==':' \
							|| XML_ISCombiningChar(c) || XML_ISExtender(c) )
#define NameInitial(c)	( Letter(c) || c=='_' || c==':' )

static XMLError IdentifierRun( XML * x, StateRecord *sr )
{
	int		c;
	enum { INITIAL, NAMECHAR };
	
	c = LexBuffer_NextChar(x->LexBuffer);
	if (c!=-1)
	{

		switch (sr->state)
		{
		case INITIAL:
			if ( NameInitial(c) )
				sr->state=NAMECHAR;
			else /* No match */
			{
				LexBuffer_PrevChar(x->LexBuffer);
				SM_Return(x);
			}
			break;

		case NAMECHAR:
			if ( !NameChar(c) )
			{
				DStr	**id = sr->param1.t_ppDStr;
				LexBuffer_PrevChar(x->LexBuffer);
				LexBuffer_Copy(x->LexBuffer, id);
				LexBuffer_Accept(x->LexBuffer);
				SM_Return(x); /* Go back to calling state */
			}
			break;
		}
	}
	return eXML_OK;
}

/* Attribute parse
******************************************************************************/

/* For building an attribute list */
typedef struct
{
	DStr		*name;
	DStr		*value;
	AttribList	*list;
	AttribList	*endlist;
} AttribWork;

static XMLError AttributeInit( XML *x, StateRecord *sr )
{
	sr->work = malloc( sizeof (AttribWork) );
	if ( sr->work )
	{
		/* ok */
		AttribWork	*w = sr->work;
		w->name=NULL;
		w->value=NULL;
		w->list=NULL;
		w->endlist=NULL;
	}
	else
		return eXML_OutOfMemory;/* memory */
	return eXML_OK;
}

static XMLError AttributeRun( XML *x, StateRecord *sr )
{
	XMLError	result = eXML_OK;
	enum AttribStates { PRE_WHITE, NAME, CHECK_NAME, EQUAL_OP, VAL, CHECK_VAL, COLLECT };
	int		c;
	AttribList	*newattr;
	AttribWork	*w = sr->work;

	switch (sr->state)
	{
	case PRE_WHITE:
		sr->state = NAME;
		SM_Call(x, &WhiteScan ); /* Remove white space after tag name */
		break;

	case NAME:
		{
			Parameter p; p.t_ppDStr=&w->name;
			sr->state = CHECK_NAME;
			SM_Call1( x, &IdentifierScan, &p );
		}
		break;

	case CHECK_NAME:
		if ( w->name )
		{
			sr->state = EQUAL_OP;
			SM_Call(x, &WhiteScan ); /* Scan for whitespace after name */
		}
		else /* No more, out we go */
		{
			AttribList **l = sr->param1.t_ppvoid;
			*l = w->list;
			w->endlist=NULL;
			w->list=NULL;
			w->name=NULL;
			w->value=NULL;
			SM_Return(x);
		}
		break;

	case EQUAL_OP:
		c = LexBuffer_NextChar(x->LexBuffer);
		if (c!=-1)
		{
			if (c != '=') /* should be = */
			{
				LexBuffer_PrevChar(x->LexBuffer);
				NOTIFY_PARSE_ERROR(x, eXMLParseErr_ExpectingEquals);
				SM_Return(x);/* Parse Error */
			}
			else
			{
				LexBuffer_Accept(x->LexBuffer);
				sr->state = VAL;
				SM_Call(x, &WhiteScan ); /* Scan for white space after equal op */
			}
		}
		break;

	case VAL:
		{
			Parameter p; p.t_ppDStr=&w->value;
			/* Get a string */
			sr->state = CHECK_VAL;
			SM_Call1(x, &StringScan, &p );
		}
		break;

	case CHECK_VAL:
		if ( w->value )
		{
			sr->state = COLLECT;
		}
		else	/* Parse error */
		{
			NOTIFY_PARSE_ERROR(x, eXMLParseErr_ValueMalformed);
			SM_Return(x);
		}
		break;

	case COLLECT:
		newattr = malloc( sizeof (AttribList) );
		if ( newattr )
		{
			if ( w->list == NULL ) /* Should maybe use a dummy first list here */
			{
				w->list = newattr;
				w->endlist = newattr;
			}
			w->endlist->next = newattr;
			w->endlist = newattr;
			newattr->next=NULL;
			
			newattr->name=w->name; w->name=NULL;
			newattr->value=w->value; w->value=NULL;
			sr->state = PRE_WHITE;
		}
		else
			result = eXML_OutOfMemory; /* Memory */
		break;

	}
	return result;
}

static XMLError AttributeDone( XML *x, StateRecord *sr )
{
	AttribList	*al;

	if ( sr->work )
	{
		AttribWork	*w = sr->work;
		if ( w->name )
			free(w->name);
		if ( w->value )
			free(w->value);
		al = w->list;
		while (al)
		{
			AttribList	*temp = al->next;
			free(al->name);
			free(al->value);
			free(al);
			al=temp;
		}
		free( w );
	}
	return eXML_OK;
}

/* String parse
*********************************************************************************/
static XMLError StringRun( XML * x, StateRecord *sr )
{
#define QUOTE_CHAR	(sr->local1.t_int)
#define STRING		(sr->param1.t_ppDStr)
	int		c;
	enum { INIT, OPEN, CLOSE };	

	switch (sr->state)
	{
	case INIT:
		if ( *STRING == NULL )
		{
			if ( DStr_Create( STRING ) )
			{
				SM_Return(x);
				return eXML_OutOfMemory;
			}
		}
		sr->state = OPEN;
		break;

	case OPEN: /* Open quote */
		c = LexBuffer_NextChar(x->LexBuffer);
		if (c==-1)
			return eXML_OK;
		if ( c == '"' || c=='\'' )
		{
			LexBuffer_Accept(x->LexBuffer);
			sr->state = CLOSE;
			QUOTE_CHAR = c; /* This is what the closing quote will be */
		}
		else
		{
			LexBuffer_Rewind(x->LexBuffer);
			SM_Return(x); /* Parse Error */
		}
		break;

	case CLOSE: /* Search for close quote */
		c = LexBuffer_NextChar(x->LexBuffer);
		if (c==-1)
			return eXML_OK;
		if ( c == QUOTE_CHAR )
		{
			LexBuffer_PrevChar(x->LexBuffer);
			LexBuffer_Append(x->LexBuffer, STRING);
			LexBuffer_NextChar(x->LexBuffer);
			LexBuffer_Accept(x->LexBuffer);
			SM_Return(x);
		}
		else if ( c == '&' ) /* An entity reference in the string, who would have believed it */
		{
			Parameter	p1; p1.t_ppDStr = STRING;
			LexBuffer_PrevChar(x->LexBuffer);
			LexBuffer_Append(x->LexBuffer, STRING);
			LexBuffer_Accept(x->LexBuffer);
			SM_Call1( x, &AmpersandScan, &p1 ); 
		}
		break;
	}
	return eXML_OK;
#undef QUOTE_CHAR
}
/* Character Reference entity 
********************************************************************************************/
static XMLError CharacterReferenceRun( XML *x, StateRecord *sr)
{
	XMLError	result = eXML_OK;
	int	c;
	do
		c = LexBuffer_NextChar(x->LexBuffer);
	while ( c!=-1 && c!=';');
	if ( c!=-1 )
	{
		DStr	*num;
		XMLChar *endptr;
		DStr	**crep = sr->param1.t_ppDStr;
		LexBuffer_PrevChar(x->LexBuffer);
		if ( LexBuffer_Gap(x->LexBuffer) <= 10 )	/* More than 10 is more than 4 billion in decimal, and is far too big */
		{
			unsigned long i;
			LexBuffer_Copy(x->LexBuffer, &num);
			if (num)
			{
				i = XML_STRTOL( DStr_CStr(num), &endptr, sr->param2.t_int);
				if ( *endptr == '\0' ) 
				{
					DStr_AppendChar( crep, (XMLChar)i );
					LexBuffer_Accept(x->LexBuffer);
				}
				else 
					NOTIFY_PARSE_ERROR(x, eXMLParseErr_MalformedInteger);/* Parse error */
				DStr_Destroy(&num);
			}
			else
				NOTIFY_PARSE_ERROR(x, eXMLParseErr_MalformedInteger);/* Parse error */
		}
		else
			NOTIFY_PARSE_ERROR(x, eXMLParseErr_OutOfRange);/* Parse error */;
		SM_Return(x);
	}
	return result;
}


/* white space eliminator
*******************************************/
#define IS_XML_WHITE(c)	( (c)==0x20 || (c)==0x9 || (c)==0x0d || (c)==0x0a )
static XMLError WhiteRun( XML *x, StateRecord *sr )
{
	int c;
	/* Eliminate white space */
	do
		c = LexBuffer_NextChar(x->LexBuffer);
	while ( c != -1 && IS_XML_WHITE(c) ); /* While not run out of chars, and is space */
	if (c!=-1)
	{
		LexBuffer_PrevChar(x->LexBuffer); /* Back up to last space */
		LexBuffer_Accept(x->LexBuffer);
		SM_Return(x);
	}
	return eXML_OK;	
}

/* Comment parser 
*****************************************************************************/
/* The sequence -- is not allowed within a comment */

static XMLError	CommentRun( XML *x, StateRecord *sr )
{
	enum { INITIAL_DASH, DASH1, DASH2, CLOSE };
	int	c;

	c = LexBuffer_NextChar(x->LexBuffer);
	if (c==-1)
		return eXML_OK;
	switch (sr->state)
	{
		case INITIAL_DASH:
			if ( c=='-' )
			{
				sr->state=DASH1;
				LexBuffer_Accept(x->LexBuffer);
			}
			else
			{
				NOTIFY_PARSE_ERROR(x, eXMLParseErr_MalformedComment);
				SM_Return(x);
			}
			break;

		case DASH1:	/* Look for first end dash */
			if ( c=='-' )
				sr->state=DASH2;
			break;
		case DASH2: /* Look for second end dash */
			if ( c=='-') 
				sr->state=CLOSE;
			else
				sr->state=DASH1;
			break;
		case CLOSE: /* Verify this is a proper ending, but note that -- is ilegal in the comment stream */
			if ( c=='>' )
			{
				DStr *comment;
				/* Ok then */
				LexBuffer_Move(x->LexBuffer, -3);
				LexBuffer_Copy(x->LexBuffer, &comment );
				NOTIFY_COMMENT(x, comment );
				DStr_Destroy(&comment);
				LexBuffer_Move(x->LexBuffer, 2);
				LexBuffer_Accept(x->LexBuffer);
			}
			else /* A "--" string was found mid stream, this is not allowed */
			{
				NOTIFY_PARSE_ERROR(x, eXMLParseErr_DoubleDashInCommentStream);
			}
			SM_Return(x);
			break;
	}
	return eXML_OK;
}
/* Process doctype - ignore it 
***********************************************************************************/

static XMLError	DoctypeRun( XML *x, StateRecord *sr )
{
	enum { INIT, DOCTYPE, CLOSE_ANGLE, CLOSE_SQUARE };
	int c = LexBuffer_NextChar(x->LexBuffer);
	const XMLChar * const doctypesig = XML_TEXT("DOCTYPE");

	if ( c==-1)
		return eXML_OK;
	switch ( sr->state )
	{
	case INIT:
		sr->param1.t_cpXMLChar = doctypesig+1;
		sr->state=DOCTYPE;
		break;

	case DOCTYPE:
		if ( c==*sr->param1.t_cpXMLChar )
		{
			sr->param1.t_cpXMLChar++;
			if ( *sr->param1.t_cpXMLChar == '\0' )
			{
				sr->state = CLOSE_ANGLE;
			}
		}
		else
		{
			NOTIFY_PARSE_ERROR(x, eXMLParseErr_MalformedDOCTYPE);
			SM_Return(x);
		}
		break;

	case CLOSE_ANGLE:
		if ( c=='>' )
		{
			LexBuffer_PrevChar(x->LexBuffer);
			SM_Return(x);
		}
		if ( c=='[' )
			sr->state=CLOSE_SQUARE;
		break;
	case CLOSE_SQUARE:
		if ( c==']' )
			sr->state=CLOSE_ANGLE;
		break;
	}
	return eXML_OK;
}
/* Process PI tags
***********************************************************************************************************/

static XMLError	PITagRun( XML *x, StateRecord *sr )
{
	TagWork *w = sr->work;
	int c;
	enum { NAME_AND_ATTRIBUTE, NOTIFY };

	switch (sr->state)
	{
	case NAME_AND_ATTRIBUTE:
		{
			Parameter p1, p2;
			p1.t_ppDStr=&w->name; p2.t_ppvoid=&w->attribs;
			SM_Call2(x, &NameAndAttributeScan, &p1, &p2 );
			sr->state=NOTIFY;
		}
		break;

	case NOTIFY:
		c = LexBuffer_NextChar(x->LexBuffer);
		if (c!=-1)
		{
			if ( c=='?' )
			{
				LexBuffer_Accept(x->LexBuffer);
				if ( w->name )
					NOTIFY_PI_TAG(x, w->name, w->attribs );
			}
			else
				NOTIFY_PARSE_ERROR(x, eXMLParseErr_MalformedPITag );
			SM_Return(x);
		}
		break;

	}
	return eXML_OK;
}
/* CDATA section scan
**************************************************************************************************************/
/* 
Start: "<![CDATA[" 
End: "]]>"
*/
static XMLError	CDATARun( XML *x, StateRecord *sr )
{
#define STR_STATE	(sr->local1.t_cpXMLChar)
#define TEXTCOUNT	(sr->local2.t_usigned_long)
	XMLChar const * const opensig = XML_TEXT("<![CDATA[");
	const XMLChar * const closesig = XML_TEXT("]]>");
	int		c;
	DStr	*dest;
	enum	{ INIT, OPEN, CLOSE };

	switch (sr->state)
	{
	case INIT:
		STR_STATE = opensig+3; /* 3 chars already matched when we got here */
		TEXTCOUNT = 0;	/* Count how many characters output */
		sr->state = OPEN;
		break;

	case OPEN:
		c = LexBuffer_NextChar(x->LexBuffer);
		if (c!=-1)
		{
			if (c==*STR_STATE)
			{
				STR_STATE++;
				if ( *STR_STATE == 0) /* End of sig? */
				{
					sr->state = CLOSE;
					STR_STATE = closesig;
					LexBuffer_Accept(x->LexBuffer);
				}
			}
			else
			{
				NOTIFY_PARSE_ERROR(x, eXMLParseErr_MalformedCDATASection);
				SM_Return(x);
			}
		}
		break;

	case CLOSE:
		c = LexBuffer_NextChar(x->LexBuffer);
		if (c!=-1)
		{
			TEXTCOUNT++;
			if ( TEXTCOUNT == x->options.MaxTextChunk )
			{
				LexBuffer_Move(x->LexBuffer, -(int)sizeof closesig); /* Go back incase we accidentally capture a ]]> */
				LexBuffer_Copy(x->LexBuffer, &dest);
				if ( dest )
				{
					NOTIFY_TEXT(x, dest );
					DStr_Destroy(&dest);
				}
				LexBuffer_Accept(x->LexBuffer);
				LexBuffer_Move(x->LexBuffer, sizeof closesig); /* Move back to where we were */
				TEXTCOUNT=0;
			}
			if (c==*STR_STATE)
			{
				STR_STATE++;
				if ( *STR_STATE == 0) /* End of sig? */
				{
					
					LexBuffer_Move(x->LexBuffer, -3);
					LexBuffer_Copy(x->LexBuffer, &dest);
					if ( dest )
					{
						NOTIFY_TEXT(x, dest );
						DStr_Destroy(&dest);
					}
					LexBuffer_Move(x->LexBuffer, 2); /* Leave the > for the caller */
					LexBuffer_Accept(x->LexBuffer);
					SM_Return(x);
				}
			}
			else
			{
				STR_STATE = closesig;
			}
		}
		break;
	}
	return eXML_OK;
#undef TEXTCOUNT
}
