
#include <stdlib.h>
#include <assert.h>

#include "xml.h"

void SM_Call( XML *x, ScannerDescriptor* sd )
{
	Parameter	p1, p2;
	SM_Call2( x, sd, &p1, &p2 );
}

void SM_Call1( XML *x, ScannerDescriptor* sd, Parameter *param1 )
{
	Parameter	p2;
	SM_Call2( x, sd, param1, &p2 );
}

void SM_Call2( XML *x, ScannerDescriptor* sd, Parameter *param1, Parameter *param2)
{
	StateMachine	*s = &x->StateMachine;
	StateRecord		*sr;
	
	assert( s->top > 0 ); /* Test stack overflow */
	s->top--;
	sr = &s->stack[ s->top ];
	sr->scanner = sd;
	sr->state = 0;
	sr->param1 = *param1;
	sr->param2 = *param2;
	sr->work=NULL;
	if (sr->scanner->init )
		sr->scanner->init( x, sr );
}


void SM_Return( XML *x )
{
	StateMachine	*s = &x->StateMachine;
	StateRecord		*sr = &s->stack[ s->top ];

	if ( sr->scanner->done )
		sr->scanner->done( x, sr );
	s->top++;
}

XMLError SM_Run( XML *x )
{
	StateMachine	*s = &x->StateMachine;
	StateRecord		*sr = &s->stack[ s->top ];

	return s->stack[ s->top ].scanner->run( x, sr );
}

XMLError SM_Init( XML *x )
{
	StateMachine	*sm = &x->StateMachine;
	x->StateMachine.top = STATE_STACK_DEPTH;
	return eXML_OK;
}

XMLError SM_Done( XML *x )
{
	while (x->StateMachine.top < STATE_STACK_DEPTH)
	{
		StateRecord	*sr = &x->StateMachine.stack[ x->StateMachine.top ];
		if ( sr->scanner->done )
		{
			sr->scanner->done( x, sr );
		}
		x->StateMachine.top++;
	}
	return eXML_OK;
}