#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "xml.h"
#include "LexBuffer.h"



XMLError XML_Create( XML** xml )
{
	XML*	x;
	XMLError		ret = eXML_OK;

	assert( xml != NULL );
	*xml = malloc( sizeof (XML) );
	if ( *xml ) 
	{
		x = *xml;
		
		SM_Init( x );	
		x->UserData=NULL;
		x->TextNotify=NULL;
		x->StartTagNotify=NULL;
		x->EndTagNotify=NULL;
		x->CommentNotify=NULL;
		x->PITagNotify=NULL;
		x->ParseErrorNotify=NULL;
		x->LexBuffer = LexBuffer_Create();
		x->options.MaxTextChunk = INITIAL_MAX_TEXT_CHUNK;
		if ( x->LexBuffer )
			SM_Call( x, &TextScan );
		else
			ret = eXML_FailedToCreateLexBuffer;

	}
	else
	{
		ret = eXML_OutOfMemory;
	}


	if ( ret )	/* If error cleanup */
	{
		XML_Destroy( xml );		
	}

	return ret;
}

XMLError XML_Destroy( XML** xml )
{
	XMLError ret = eXML_OK;

	assert( xml != NULL );

	if ( *xml )
	{
		XML *x = *xml;
		if ( x->LexBuffer )
			LexBuffer_Destroy( x->LexBuffer );
		SM_Done( x );
		free( *xml );
		*xml = NULL;
	}
	return ret;
}

void XML_SetUserData( XML* x, void *user_data )
{
	x->UserData = user_data;
}

void* XML_GetUserData( XML* x )
{
	return x->UserData;
}

XMLError XML_SetTextHandler( XML *x, TextNotify* n )
{
	x->TextNotify = n;
	return eXML_OK;
}

XMLError XML_SetStartTagHandler( XML *x, StartTagNotify* n )
{
	x->StartTagNotify = n;
	return eXML_OK;
}

XMLError XML_SetEndTagHandler( XML *x, EndTagNotify* n )
{
	x->EndTagNotify = n;
	return eXML_OK;
}

XMLError XML_SetCommentHandler( XML* x, CommentNotify* n )
{
	x->CommentNotify = n;
	return eXML_OK;
}

XMLError XML_SetPITagHandler( XML *x, PITagNotify* n )
{
	x->PITagNotify = n;
	return eXML_OK;
}

XMLError XML_SetParseErrorHandler( XML* x, ParseErrorNotify* n )
{
	x->ParseErrorNotify = n;
	return eXML_OK;
}

XMLError XML_SetMaxTextChunk( XML *xml, size_t size )
{
	assert( xml != NULL );
	xml->options.MaxTextChunk = size;
	return eXML_OK;
}

XMLError XML_ParseBlock( XML *xml, XMLChar *buf, size_t n )
{
	XMLError result;

	if ( n && buf )
	{
		if ( LexBuffer_AddData(xml->LexBuffer, buf, n) )
			return 1;
		
	}

	for ( result=eXML_OK; !LexBuffer_IsEmpty( xml->LexBuffer ) && result == eXML_OK;) 
			result = SM_Run( xml );
	
	return result;
}

