#ifndef LEXBUFFER_H
#define LEXBUFFER_H

#include "xmltypes.h"
#include "DStr.h"

#define LEX_TAB_WIDTH	4
#define	LEX_BLOCK_SIZE	100

typedef struct LexBlock LexBlock;
struct LexBlock {
	int				n;					/* Data count */
	XMLChar			d[LEX_BLOCK_SIZE];	/* Data */
	LexBlock		*prev;
	LexBlock		*next;
};



typedef struct LexBuffer LexBuffer;
struct LexBuffer {
	LexBlock*		startBlock;		/* Start of LexBuffer data pointers */
	int				start;
	LexBlock*		forwardBlock;	/* Scan forward data pointers */
	int				forward;
	LexBlock*		endBlock;		/* End data pointers */
	int				gap;

	/* For counting source lines and columns */
	unsigned long	lineNo;
	unsigned long	colNo;
	size_t			tabSize;
};



extern	LexBuffer	*LexBuffer_Create(void);
extern	int	LexBuffer_Destroy(LexBuffer*);
extern	int	LexBuffer_SetTabSize(LexBuffer*, size_t);
extern	int	LexBuffer_AddData(LexBuffer*, XMLChar*, size_t);
extern	tInt LexBuffer_NextChar(LexBuffer*);
extern	tInt LexBuffer_PrevChar(LexBuffer*);
extern	int LexBuffer_Move(LexBuffer *lex, int delta);
extern	int	LexBuffer_Accept(LexBuffer*);
extern	int	LexBuffer_Rewind(LexBuffer*);
extern	int LexBuffer_Copy(LexBuffer *, DStr **);
extern	int LexBuffer_Append(LexBuffer *, DStr **);
extern	int LexBuffer_IsEmpty(LexBuffer *);
extern	int	LexBuffer_Gap(LexBuffer *);




#endif