#ifndef STATEMACHINE_H
#define STATEMACHINE_H

#include "xmltypes.h"
#include "xmlerror.h"
#include "DStr.h"

#define STATE_STACK_DEPTH	10

typedef struct AttribList {
	DStr				*name;
	DStr				*value;
	struct AttribList	*next;
} AttribList;


typedef struct XML XML;
typedef struct StateRecord StateRecord;
typedef struct ScannerDescriptor ScannerDescriptor;
typedef struct StateMachine StateMachine;
typedef XMLError (ScannerFn)( XML *, StateRecord *);


struct ScannerDescriptor {
	ScannerFn*		init;
	ScannerFn*		run;
	ScannerFn*		done;
};

typedef union Parameter Parameter;
union Parameter
{
	void *			t_pvoid;
	void *			t_ppvoid;
	XMLChar *		t_pXMLChar;
	const XMLChar *	t_cpXMLChar;
	XMLChar **		t_ppXMLChar;
	DStr *			t_pDStr;
	DStr **			t_ppDStr;
	int				t_int;
	int *			t_pint;
	unsigned long	t_usigned_long;
};

struct StateRecord {
	ScannerDescriptor*		scanner;	/* The current state */
	int						state;		/* Prvate state integer */
	void					*work;		/* Private work area */
	Parameter				param1;		/* Param 1 in */
	Parameter				param2;		/* Param 2 in */
	Parameter				local1;
	Parameter				local2;
};

struct StateMachine {
	int					top;
	StateRecord			stack[STATE_STACK_DEPTH];
};


typedef void (TextNotify)( void	*UserData, DStr * text );
typedef void (StartTagNotify)( void	*UserData, DStr* name, AttribList* attribs );
typedef void (EndTagNotify)( void *UserData, DStr *name );
typedef void (CommentNotify)( void *UserData, DStr *comment );
typedef void (PITagNotify)( void *UserData, DStr *name, AttribList* attribs );
typedef void (ParseErrorNotify)( void *UserData, unsigned long line, unsigned long column, XMLError err );



extern	ScannerDescriptor	TextScan;

extern	void SM_Call( XML *, ScannerDescriptor* );
extern	void SM_Call1( XML *, ScannerDescriptor*, Parameter * );
extern	void SM_Call2( XML *, ScannerDescriptor*, Parameter *, Parameter *);
extern	void SM_Return( XML * );
extern	XMLError SM_Run( XML * );
extern	XMLError SM_Init( XML * );
extern	XMLError SM_Done( XML * );

#endif /* #ifndef STATEMACHINE_H */
