#ifndef CLASSMAP_H
#define CLASSMAP_H

#define BITS_PER_MEL	8	/* This MUST be equal to the number of bits in the typedef Mel */
typedef unsigned char Mel;	/* Map-Element, the fundamental type that is manipulated in the map */
typedef unsigned char PageMap;


#define PAGE_MAX		256
#define CHARMAP_MAX		( 65536 / BITS_PER_MEL )
#define CHARPMAP_SIZE	( 65536 / 8 )
#define BYTES_PER_PAGE	32
#define MELS_PER_PAGE	(PAGE_MAX / BITS_PER_MEL)

/* Optimise in specific cases, becuase we can't precompute log2(BITS_PER_MEL) using the preprocessor */
#if BITS_PER_MEL == 8 

#define SETBIT(target, bit)		((target)[(bit)>>3]|=1<<((bit)&7))
#define GETBIT(target, bit)		((target)[(bit)>>3] & (1<<((bit)&7)))

#else

#define SETBIT(target, bit)		((target)[(bit)/BITS_PER_MEL]|=1<<(bit)%BITS_PER_MEL)
#define GETBIT(target, bit)		((target)[(bit)/BITS_PER_MEL] & ((bit)%BITS_PER_MEL))

#endif /* #if BITS_PER_MEL == 8 */

#define XML_ISMEMBER( cmap, pagemap, code )		GETBIT( &cmap[ MELS_PER_PAGE * pagemap[ code >> 8 ] ], code & 0xff )

#define XML_ISBaseChar( code )					XML_ISMEMBER( Shared_Map, BaseChar_Page, code )
#define XML_ISIdeogaphic( code )				XML_ISMEMBER( Shared_Map, Ideographic_Page, code )
#define XML_ISCombiningChar( code )				XML_ISMEMBER( Shared_Map, CombiningChar_Page, code )
#define XML_ISDigit( code )						XML_ISMEMBER( Shared_Map, Digit_Page, code )
#define XML_ISExtender( code )					XML_ISMEMBER( Shared_Map, Extender_Page, code )




extern const PageMap BaseChar_Page[256];
extern const PageMap Ideographic_Page[256];
extern const PageMap CombiningChar_Page[256];
extern const PageMap Digit_Page[256];
extern const PageMap Extender_Page[256];
extern Mel const Shared_Map[];

#endif