
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "xml.h"
#include "DStr.h"

#ifdef UNICODE
#define XML_PRINTF	wprintf
#else
#define XML_PRINTF	printf
#endif

void start_tag_notify( void *UserData, DStr* name, AttribList* attribs )
{
	unsigned int *uf = UserData;

	XML_PRINTF(XML_TEXT("(%s"), DStr_CStr(name) );
	while ( attribs )
	{
		XML_PRINTF(XML_TEXT(" %s=\"%s\""), DStr_CStr(attribs->name), DStr_CStr(attribs->value) );
		attribs = attribs->next;
	}
	XML_PRINTF(XML_TEXT(") {"));
	if ( *uf & 01 )
		XML_PRINTF(XML_TEXT("\n"));
}

void text_notify( void *UserData, DStr * text )
{
	unsigned int *uf = UserData;

	XML_PRINTF(XML_TEXT("%s"), DStr_CStr(text) );
	if ( *uf & 02 )
		XML_PRINTF(XML_TEXT("\n"));
}

void end_tag_notify( void *UserData, DStr *name )
{
	unsigned int *uf = UserData;

	XML_PRINTF(XML_TEXT("} (/%s)"), DStr_CStr(name) );
	if ( *uf & 04 )
		XML_PRINTF(XML_TEXT("\n"));
}

void comment_notify( void *UserData, DStr *comment )
{
	XML_PRINTF(XML_TEXT("/* %s */"), DStr_CStr(comment) );
}

void pi_tag_notify( void *UserData, DStr* name, AttribList* attribs )
{
	unsigned int *uf = UserData;

	XML_PRINTF(XML_TEXT("#%s"), DStr_CStr(name) );
	while ( attribs )
	{
		XML_PRINTF(XML_TEXT(" %s=\"%s\""), DStr_CStr(attribs->name), DStr_CStr(attribs->value) );
		attribs = attribs->next;
	}
	XML_PRINTF(XML_TEXT("#"));
	if ( *uf & 01 )
		XML_PRINTF(XML_TEXT("\n"));
}

void parse_error_notify( void *UserData, unsigned long line, unsigned long column, XMLError err)
{
	printf("\n[PARSE ERROR near Line %lu, column %lu, `%s`]\n", line, column, XML_GetErrStr(err));
}

int main( int argc, char* argv[] ) 
{
	FILE*	in;
	XMLChar	buf[1000];
	size_t	n;
	XML*	xml;
	int		err = 0;
	unsigned int newline_flags=0;

	XML_Create( &xml );

	XML_SetUserData( xml, &newline_flags );

	/* Comment these lines out to see different output */
	XML_SetStartTagHandler( xml, start_tag_notify ); newline_flags |= 01;
	XML_SetEndTagHandler( xml, end_tag_notify ); newline_flags |= 04;
	XML_SetTextHandler( xml, text_notify ); newline_flags = 0;
	XML_SetCommentHandler( xml, comment_notify );
	XML_SetPITagHandler( xml, pi_tag_notify );
	XML_SetParseErrorHandler( xml, parse_error_notify );

	/* Set maximum text chunck to return in notify fn */
	XML_SetMaxTextChunk( xml, 10); 

	printf("Opening %s\n", argv[1]);
	in = fopen( argv[1], "rb" );
	if ( in == NULL )
	{
		printf("%s: <xml-file>\n", argv[0]);
		abort();
	}

	n = fread(buf, sizeof (XMLChar), sizeof buf / sizeof buf[0], in);
	while ( n > 0 )
	{
		err = XML_ParseBlock( xml, buf, n );
		if ( err )
			break;
		n = fread(buf, sizeof (XMLChar), sizeof buf / sizeof buf[0], in);
	}
	
	if ( ! err )
		err = XML_ParseBlock( xml, NULL, 0 );

	XML_Destroy( &xml );

#if !defined(NDEBUG)	
	_CrtDumpMemoryLeaks();
#endif

	printf("Finished with exit code = %d\n", err);
	system("pause");

	return	EXIT_SUCCESS;
	(void)	argc;
}